/*

    File: CModel.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "../m_system3d/error.h"
#include "../m_foundation/const.h"
#include "cmodel.h"

void CModel::Reset(void)
{
	vtxCount = 0;
	vtx = NULL;

	planeCount = 0;
	edgeCount = 0;
	triCount = 0;

	plane = NULL;
	edge = NULL;
	tri = NULL;
}

void CModel::Unload(void)
{
	if (vtx) free(vtx);
	if (plane) free(plane);
	if (edge) free(edge);
	if (tri) free(tri);

	Reset();
}

bool CModel::Load(const char *filename)
{
	int file_vtxCount;
	int file_triCount;
	int t,v;

	//CUSTOM MDL FILE FORMAT

	strcpy(title, filename);

	char sig[4];

	FILE *File = fopen(filename, "rb");
	if (File == NULL)
	{
		Error("%s: unable to open model file", filename);
		return false;
	}

	fread(sig, 1, 4, File);
	if (memcmp(sig, "MMDL", 4))
	{
		Error("%s: bad signature on file", filename);
		return false;
	}
	fread(&file_vtxCount, sizeof(int), 1, File);
	fread(&file_triCount, sizeof(int), 1, File);

	if (file_vtxCount < 0)
	{
		Error("%s: cmodel has bad vertex count: %d", filename, file_vtxCount);
		fclose(File);
		return false;
	}
	if (file_triCount < 0)
	{
		Error("%s: cmodel has bad plane count: %d", filename, file_triCount);
		fclose(File);
		return false;
	}

	//allocate verticies
	if ((vtx = (vector_t *)malloc(sizeof(vector_t) * file_vtxCount))==NULL)
	{
		Error("%s: unable to allocate vertex array", filename);
		fclose(File);
		return false;
	}
	//read verticies
	for (v = 0; v < file_vtxCount; v++)
	{
		//always read float[]'s
		float tvtx[3];
		if (fread(tvtx, sizeof(float)*3, 1, File) != (size_t)(1))
		{
			Error("%s: unable to read vertex array", filename);
			free(vtx);
			fclose(File);
			return false;
		}
		vtx[v].x = (scalar_t)tvtx[0];
		vtx[v].y = (scalar_t)tvtx[1];
		vtx[v].z = (scalar_t)tvtx[2];
	}
	//skip texture coordinates
	if (fseek(File, sizeof(float)*file_vtxCount*2, SEEK_CUR))
	{
		Error("%s: unable to skip texture coordinate array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	//allocate triangles
	if ((tri = (int *)malloc(sizeof(int) * 3 * file_triCount))==NULL)
	{
		Error("%s: unable to allocate tri array", filename);
		free(vtx);
		fclose(File);
		return false;
	}
	//read triangles
	for (t = 0; t < file_triCount; t++)
	{
		int swap;
		if (fread(&tri[3*t], sizeof(int)*3, 1, File) != 1)
		{
			Error("%s: unable to read tri array", filename);
			free(vtx);
			free(tri);
			fclose(File);
			return false;
		}
		swap = tri[3*t];
		tri[3*t] = tri[3*t+2];
		tri[3*t+2] = swap;
	}

	fclose(File);

	if ((plane = (plane_t *)malloc(sizeof(plane_t) * file_triCount))==NULL)
	{
		Error("%s: unable to allocate plane array", filename);
		free(tri);
		free(vtx);
		fclose(File);
		return false;
	}

	//calc edgeCount
	edgeCount = 0;
	for (int a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t]) count++;
				}
			}

			if (count == 2) edgeCount++;
		}
	}
	//allocate edge buffer
	if ((edge = (edge_t *)malloc(edgeCount * sizeof(edge_t)))==NULL)
	{
		Error("%s: unable to allocate edge array", filename);
		free(vtx);
		free(tri);
		free(plane);
		return false;
	}
	//create edges
	int l = 0;
	for (a = 0; a < file_triCount-1; a++)
	{
		for (int b = a+1; b < file_triCount; b++)
		{
			edge_t e;
			int count = 0;

			//if a and b share two verticies - create an edge
			for (v = 0; v < 3; v++)	//verticies of tri a
			{
				for (t = 0; t < 3; t++)	//verticies of tri b
				{
					if (tri[3*a+v] == tri[3*b+t])
					{
						if (count < 2)
						{
							//store triangles
							e.t[0] = a;
							e.t[1] = b;
							//store edge
							e.v[count] = tri[3*a+v];
						}
						count++;
					}
				}
			}

			if (count == 2)
			{
				memcpy(&edge[l], &e, sizeof(edge_t));

//				edge[l].normal = plane[e.t[0]].normal + plane[e.t[1]].normal;
//				edge[l].normal.Normalize();
				
				l++;
			}
		}
	}

	vtxCount = file_vtxCount;
	planeCount = file_triCount;
	triCount = file_triCount;

	CalcPlanes();
	CalcBounds();
	CalcVolume();

	return true;
}

void CModel::CalcBounds(void)
{
	int v;

	//calc bbox
	VectorSet(&bbox.min, fInf,fInf,fInf);
	VectorSet(&bbox.max, -fInf,-fInf,-fInf);
	radius = -1;
	for (v = 0; v < vtxCount; v++)
	{
		//stretch bbox
		BoxStretch(&bbox, &vtx[v]);

		//stretch radius
		scalar_t l = VectorLength(&vtx[v]);
		if (l > radius) radius = l;
	}
}

void CModel::CalcPlanes(void)
{
	int t,v;
	for (v = 0; v < planeCount; v++)
	{
		t = v*3;
		VectorPlaneUnitNormal(&vtx[tri[t+2]], &vtx[tri[t+1]], &vtx[tri[t]], &plane[v].normal);
		PlaneConstruct(&vtx[tri[t]], &plane[v].normal, &plane[v]);
		
		//plane[v].pos = (vtx[tri[t]] + vtx[tri[t+1]] + vtx[tri[t+2]]) * (1.f/3.f);
		//plane[v].normal = PlaneUnitNormal(vtx[tri[t+2]], vtx[tri[t+1]], vtx[tri[t]]);
	}
}

void CModel::Transform(const quat_t *r, const vector_t *s, const vector_t *x)
{
	int i;

	for (i = 0; i < vtxCount; i++)
	{
		vtx[i].x *= s->x;
		vtx[i].y *= s->y;
		vtx[i].z *= s->z;
		QuatRotateVector(&vtx[i], r, &vtx[i]);
		VectorAdd(&vtx[i], x, &vtx[i]);
		//vtx[i] = RotateVector(ScaleVector(vtx[i],s), r) + x;
	}

//	for (i = 0; i < edgeCount; i++)
//		edge[i].normal = RotateVector(edge[i].normal, r);

	CalcPlanes();
	CalcBounds();

	//quick mod of volume - need not calc it again
	volume *= s->x * s->y * s->z;
}

bool CModel::Copy(CModel *src)
{
	Unload();	//make sure we're clear

	vtxCount = src->vtxCount;
	if ((vtx = (vector_t *)malloc(sizeof(vector_t) * vtxCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(vtx, src->vtx, sizeof(vector_t) * vtxCount);

	planeCount = src->planeCount;
	if ((plane = (plane_t *)malloc(sizeof(plane_t) * planeCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(plane, src->plane, sizeof(plane_t) * planeCount);

	edgeCount = src->edgeCount;
	if ((edge = (edge_t *)malloc(edgeCount * sizeof(edge_t)))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(edge, src->edge, edgeCount * sizeof(edge_t));

	triCount = src->triCount;
	if ((tri = (int *)malloc(sizeof(int) * 3 * triCount))==NULL)
	{
		Unload();
		return false;
	}
	memcpy(tri, src->tri, sizeof(int) * 3 * triCount);

	return true;
}

void CModel::CalcVolume(void)
{
	//calculate volume based on tris and vtxs
	volume = 0;
	for (int t = 0; t < triCount*3; t+=3)
	{
		vector_t *pv[3] = { &vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]] };

		volume += pv[0]->x * pv[1]->y * pv[2]->z;
		volume += pv[0]->y * pv[1]->z * pv[2]->x;
		volume += pv[0]->z * pv[1]->x * pv[2]->y;
		volume -= pv[2]->x * pv[1]->y * pv[0]->z;
		volume -= pv[2]->y * pv[1]->z * pv[0]->x;
		volume -= pv[2]->z * pv[1]->x * pv[0]->y;
	}
	volume /= 6.f;
}

void CModel::CalcInertia(const vector_t *scale, vector_t *vres)
{
	scalar_t scale_prod = scale->x * scale->y * scale->z;
	vector_t scale_sq;
	VectorSet(&scale_sq, scale->x * scale->x, scale->y * scale->y, scale->z * scale->z);

	//calculate inertia based on tris and vtxs
	vector_t inertia;
	VectorCopy(&vectorZero, &inertia);

	for (int t = 0; t < triCount*3; t+=3)
	{
		vector_t *pv[3] = { &vtx[tri[t]], &vtx[tri[t+1]], &vtx[tri[t+2]] };

/*
[ || || || ]
[          ]
[ v1 v2 v3 ]
[          ]
[ || || || ]

[ x1 x2 x3 ]
[          ]
[ y1 y2 y3 ]
[          ]
[ z1 z2 z3 ]
*/

		scalar_t det = 0;
		det += pv[0]->x * pv[1]->y * pv[2]->z;
		det += pv[0]->y * pv[1]->z * pv[2]->x;
		det += pv[0]->z * pv[1]->x * pv[2]->y;
		det -= pv[2]->x * pv[1]->y * pv[0]->z;
		det -= pv[2]->y * pv[1]->z * pv[0]->x;
		det -= pv[2]->z * pv[1]->x * pv[0]->y;

		scalar_t v0term =
			pv[0]->x * pv[0]->x +
			pv[0]->x * pv[1]->x +
			pv[0]->x * pv[2]->x +
			pv[1]->x * pv[1]->x +
			pv[1]->x * pv[2]->x +
			pv[2]->x * pv[2]->x;

		scalar_t v1term =
			pv[0]->y * pv[0]->y +
			pv[0]->y * pv[1]->y +
			pv[0]->y * pv[2]->y +
			pv[1]->y * pv[1]->y +
			pv[1]->y * pv[2]->y +
			pv[2]->y * pv[2]->y;

		scalar_t v2term =
			pv[0]->z * pv[0]->z +
			pv[0]->z * pv[1]->z +
			pv[0]->z * pv[2]->z +
			pv[1]->z * pv[1]->z +
			pv[1]->z * pv[2]->z +
			pv[2]->z * pv[2]->z;

		inertia.x += det * (v1term * scale_sq.y + v2term * scale_sq.z);
		inertia.y += det * (v0term * scale_sq.x + v2term * scale_sq.z);
		inertia.z += det * (v0term * scale_sq.x + v1term * scale_sq.y);
	}

	VectorScale(&inertia, (scale->x * scale->y * scale->z / 60.f), vres);
}

	//returns the greatest distance from any plane in the cmodel to the vector specified
int CModel::PointSurfaceDepth(const vector_t *vec, scalar_t *bestdist)
{
	*bestdist = -fInf;
	int bestplane = -1;

	for (int p = 0; p < planeCount; p++)
	{
		scalar_t dist = PlaneVectorDistance(&plane[p], vec);
		if (dist > *bestdist)
		{
			*bestdist = dist;
			bestplane = p;
		}
	}

	return bestplane;
}

scalar_t CModel::Traceline(const line_t &line, vector_t *res) const {
	vector_t src = line.pos;
	vector_t dst = line.pos + line.dir;
	scalar_t resFrac = 1;
	//for all tris
	//	calc the intersection point on the tri's plane
	//test the b.c.c values to ensure intersection

	//for the time being these two structures are parallel
	assert(planeCount == triCount);

	for (int i = 0; i < planeCount; i++) {
		scalar_t srcf = PlaneVectorDistance(&plane[i], &src);
		scalar_t dstf = PlaneVectorDistance(&plane[i], &dst);
		const int *t = tri + (3*i);

		if (srcf >= 0 && dstf >= 0) continue;
		if (srcf < 0 && dstf < 0) continue;
		
		const scalar_t FACE_STEP_EPSILON = 0.03125;

		scalar_t frac = (srcf < 0) ?
			(srcf + FACE_STEP_EPSILON) / (srcf - dstf) :
			(srcf - FACE_STEP_EPSILON) / (srcf - dstf);
		if (frac < 0) frac = 0;
		if (frac > 1) frac = 1;

		vector_t mid;
		VectorLerp(&src, &dst, frac, &mid);

		vector_t p0ToMid = mid - vtx[t[0]];
		vector_t dx	= vtx[t[1]] - vtx[t[0]];
		vector_t dy	= vtx[t[2]] - vtx[t[0]];

		vector_t dxd1 = p0ToMid * dx;
		vector_t dxd2 = p0ToMid * dy;
		vector_t d1xd2 = dx * dy;
		scalar_t div = 1.0 / VectorLengthSq(&d1xd2);
		scalar_t barycentricCoords[3];
		barycentricCoords[1] = (dxd2 % d1xd2) * div;
		barycentricCoords[2] = -(dxd1 % d1xd2) * div;
		barycentricCoords[0] = 1.0 - barycentricCoords[1] - barycentricCoords[2];

		if (barycentricCoords[0] > 0 &&
			barycentricCoords[1] > 0 &&
			barycentricCoords[2] > 0)
		{
			dst = mid;
			resFrac *= frac;
		}
	}

	if (res && resFrac < 1) *res = dst;

	return resFrac;
}
