/*

    File: Physics.cpp

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <math.h>
#include <assert.h>

#include "../m_foundation/const.h"	//Inf
#include "../m_foundation/main.h"	//console - MAKE A MORE UNIVERSAL WARNING REPORTER

#ifdef PHYSICS_PARTICLE_SYSTEM_DEBUG
#include "../m_particle/particle.h"
extern ParticleSystem *particleSystem;
#endif

#include "physics.h"		//class definitions

//#define DEBUG_PRINT

// ********************* THE RIGID BODY CLASS ********************* 

void RigidBody::CalcMassInertia(void)
{
//newschool

	if (!model) return;

	MassInv = densityInv / (model->volume * size.x * size.y * size.z);

	// ********* NOTICE ********* 
	//if a rigid body size vector changes then so does its inertia!
	vector_t inertia;

	model->CalcInertia(&size, &inertia);

	MatrixSet(&BodyInertiaInv,
		densityInv / inertia.x, 0, 0,
		0, densityInv / inertia.y, 0,
		0, 0, densityInv / inertia.z);		//need to incorperate density...
}

RigidBody::RigidBody()
{
	MassInv = 0;
	MatrixCopy(&matrixZero, &BodyInertiaInv);

	VectorCopy(&vectorZero, &x);
	QuatCopy(&quatIdentity, &q);
	VectorCopy(&vectorZero, &P);
	VectorCopy(&vectorZero, &L);

	MatrixCopy(&matrixZero, &InertiaInv);
	VectorCopy(&vectorZero, &v);
	VectorCopy(&vectorZero, &omega);

	VectorCopy(&vectorZero, &force);
	VectorCopy(&vectorZero, &torque);

	model = NULL;
	VectorSet(&size, 1,1,1);
	densityInv = 0;

	SurfaceRestitution = 1;
	
	integrate = true;	//the good ol integration flag
	collides_solid = true;
	is_solid = true;

}

RigidBody::~RigidBody()
{
	if (model) delete model;
}

void RigidBody::VelocityAtOffset(const vector_t *offset, vector_t *vres)
{
	vector_t reg;
	VectorCross(&omega, offset, &reg);
	VectorAdd(&v, &reg, vres);
	//return (v + omega * offset);
}	//'*' being the cross product

void RigidBody::ComputeInertiaInv(void)
{
	//find a better place to put this
	QuatUnit(&q, &q);
	//q.Normalize();

	matrix_t rot;
	QuatToMatrix(&q, &rot);
	//matrix_t rot = matrix_t(q);
	MatrixMul(&rot, &BodyInertiaInv, &InertiaInv);
	MatrixTranspose(&rot, &rot);
	MatrixMul(&InertiaInv, &rot, &InertiaInv);
	//InertiaInv = (rot * BodyInertiaInv) * TransposeMatrix(rot);

#if 0
	InertiaInv = (R * BodyInertiaInv) * TransposeMatrix(R);
	//orthonormalize
	vector_t vx(R.v[0][0], R.v[1][0], R.v[2][0]);
	vector_t vy(R.v[0][1], R.v[1][1], R.v[2][1]);
	vector_t vz;
	vx.normalize();
	vz = vx * vy;	vz.normalize();
	vy = vz * vx;	vy.normalize();
	R = matrix_t(vx,vy,vz);
#endif
}

void RigidBody::ComputeAuxiliary(void)		//compute auxiliary quantities - v, omega
{
	VectorScale(&P, MassInv, &v);
	//v = P * MassInv;
	MatrixMul(&InertiaInv, &L, &omega);
	//omega = InertiaInv * L;
}


// ********************* THE PHYSICS CLASS ********************* 

	//add a RigidBody to the list
void M_Physics::AddRigidBody(RigidBody *body)
{
	body->CalcMassInertia();
	body->ComputeInertiaInv();
	body->ComputeAuxiliary();

	for (int i = 0; i < rigidBodyArray.size(); i++)
	{
		if (!rigidBodyArray[i])
		{
			rigidBodyArray[i] = body;
			return;
		}
	}

	rigidBodyArray.push_back(body);
}

	//removes a RigidBody from the LinkedList - ALWAYS CALL THIS BEFORE YOU DELETE A RIGID BODY!!!
void M_Physics::RemoveRigidBody(RigidBody *remove_body)
{
	for (int i = 0; i < rigidBodyArray.size(); i++)
	{
		if (!rigidBodyArray[i]) continue;

		if (rigidBodyArray[i] == remove_body)
		{
			rigidBodyArray[i] = NULL;
//			if (i == rigidBodyArray.size() - 1)
//			{
//				do {
//					rigidBodyArraySize--;
//				} while(!rigidBodyArray[rigidBodyArraySize] && rigidBodyArraySize > 0);
//			}
		}
	}
}

bool M_Physics::Init(void /*int max*/)
{
	return true;
}

void M_Physics::Reset(void)
{
	//init variables
	planeEpsilon = 1.f;
	velEpsilon = 0.01f;
	useContactForces = 1;

	//remove all the external rigidBodies
	forceFunc = NULL;
}

void M_Physics::Kill(void)
{
	//internal
//	collisionChain.DeleteAll();
	collisionList.clear();

	Reset();
}

void M_Physics::Update(double delta_time)
{
	ApplyForces();
	UpdateSystem(delta_time);
}

void M_Physics::SetForceFunction(void (*func)(M_Physics *))
{
	//simple shell
	forceFunc = func;
}

/*
RigidBody req: just about everything, come to think of it..
child functions:	MassInv, v, omega
*/
void M_Physics::ApplyForces(void)
{
	RigidBody *body;

	for (int i = 0; i < rigidBodyArray.size(); i++)
	{
		body = rigidBodyArray[i];
		if (!body) continue;

		VectorSet(&body->force, 0,0,0);
		VectorSet(&body->torque, 0,0,0);

	}

	if (forceFunc) forceFunc(this);
}

#define CS_FREE			0
#define CS_COLLIDING	1
#define CS_PENETRATING	2

/*
RigidBody req: none
child functions:
	Integrate
	CheckCollisionState
	ResolveAllCollisions
*/
void M_Physics::UpdateSystem(double delta_time)
{
	double time, target_time;
	int collision_state;
	int mainloop;
	int penetration_accuracy;

	mainloop = 0;
	penetration_accuracy = 0;

	time = 0;
	target_time = delta_time;

	while (time != delta_time)
	{
		mainloop++;
		if (mainloop > 10)
		{
			main_console->Print("M_Physics::UpdateSystem(): Collision Frame Infinite Loop\n");
			return;
		}

		//integrate to find the dest object state based upon the source and (target_time - time)
		Integrate(target_time - time);

		//check for collision of the dest state to all other objects
		collision_state = CheckCollisionState();

		switch (collision_state)
		{
		case CS_PENETRATING:
			penetration_accuracy++;
			if (penetration_accuracy < 0)	//maximum accuracy attempted via the penetration loop
			{
				Integrate(time - target_time);				//integrate backwards / return to original state
				target_time = (time + target_time) * 0.5;	//	subdivide time: target_time = average of time and target_time
				break;										//and bail out	(we arent using penetration detection yet)
			}
//			main_console->Print("M_Physics::UpdateSystem(): Collision Accuracy Exceeded\n");
			//if we've recalculated too many times (penetration_accuracy exceeded its maximum)
			//then reset the penetration accuracy counter and continue on
			//to treat it like a regular collision
			penetration_accuracy = 0;
		case CS_COLLIDING:				//	resolve all known collisions
			ResolveAllCollisions();
			if (useContactForces)
				ComputeContactForces();		//	<- our next step to take
		case CS_FREE:
			time = target_time;			//	compute next the timeframe from 'time' to the end of our 'delta_time'
			target_time = delta_time;
			break;

		default:
			main_console->Print("M_Physics::UpdateSystem(): unknown collision state interpreted: %d\n", collision_state);
			return;
		}
	}
}

/*
RigidBody
	read:	x, P, q, L, v, force, omega, torque
	write:	x, P, q, L
child functions:
	RigidBody.ComputeInertiaInv
	RigidBody.ComputeAuxiliary
*/
void M_Physics::Integrate(double delta_time)
{
	RigidBody *body;

	for (int i = 0; i < rigidBodyArray.size(); i++)
	{
		body = rigidBodyArray[i];
		if (!body) continue;

		//if the object is static then bail out
		if (!body->integrate) continue;

		//linear integration
		VectorMultAdd(&body->x, &body->v, (scalar_t)delta_time, &body->x);
		VectorMultAdd(&body->P, &body->force, (scalar_t)delta_time, &body->P);
		//angular integration
		QuatIntegrate(&body->q, &body->omega, (scalar_t)delta_time, &body->q);
		VectorMultAdd(&body->L, &body->torque, (scalar_t)delta_time, &body->L);

		body->ComputeInertiaInv();
		body->ComputeAuxiliary();
	}
}

/*
RigidBody
	read:	model, q, size, x, collides_solid, is_solid
	write:	
child functions:
	M_Physics.TestCollision
*/
int M_Physics::CheckCollisionState(void)
{
	CModel alias[2];
	int i,j;
	int state, check;

	RigidBody *body[2];

	state = CS_FREE;
	//collisionChain.DeleteAll();		//clear the collision chain
	collisionList.clear();

	for (i = 0; i < rigidBodyArray.size()-1; i++)
	{
		bool loaded_alias_0 = false;

		body[0] = rigidBodyArray[i];
		if (!body[0]) continue;

		//if it has no collision model then dont bother testing it to anything
		if (!body[0]->model) continue;

		for (j = i+1; j < rigidBodyArray.size(); j++)
		{
			body[1] = rigidBodyArray[j];
			if (!body[1]) continue;

			//if it has no collision model then dont bother testing it to anything
			if (!body[1]->model) continue;

			if ((!body[0]->collides_solid || !body[1]->is_solid) &&
				(!body[1]->collides_solid || !body[0]->is_solid))
				continue;

			//now that all the checks are done with

			//if we havent loaded obj 0's model, then do so
			if (!loaded_alias_0)
			{
				alias[0].Copy(body[0]->model);	//copy the alias model
				alias[0].Transform(&body[0]->q, &body[0]->size, &body[0]->x);
				loaded_alias_0 = true;
			}

			alias[1].Copy(body[1]->model);	//copy the alias model
			alias[1].Transform(&body[1]->q, &body[1]->size, &body[1]->x);

			if (!BoxIntersects(&alias[0].bbox, &alias[1].bbox)) continue;

			check = TestCollision(&alias[0], &alias[1], body[0], body[1]);
			if (check > state) state = check;

			alias[1].Unload();
		}

		alias[0].Unload();
	}

	return state;
}

/*
RigidBody
	read:	x
	write:	
child functions:
*/
int M_Physics::TestCollision(CModel *m0, CModel *m1, RigidBody *o0, RigidBody *o1)
{
	CModel *m[2] = {m0, m1};
	RigidBody *o[2] = {o0, o1};

	int best_obj = -1;

	vector_t	best_point;
	vector_t	best_normal;
	vector_t	best_edge0, best_edge1;
	scalar_t	best_depth = planeEpsilon;

	int v,p;
	scalar_t depth;

	bool collided = false;
	bool vertex_face;

	int state = CS_FREE;

	vector_t o_delta;
	VectorSub(&o0->x, &o1->x, &o_delta);		//from o1 to o0

	//test object 0 vtx to object 1 plane
	for (v = 0; v < m0->vtxCount; v++)
	{
		p = m1->PointSurfaceDepth(&m0->vtx[v], &depth);
		if ((p > 0) && (depth < best_depth))
		{
			best_obj = 0;
			best_depth = depth;
			VectorCopy(&m0->vtx[v], &best_point);
			VectorCopy(&m1->plane[p].normal, &best_normal);
			collided = true;
			vertex_face = true;
		}
	}

	//test object 1 vtx to object 0 plane
	for (v = 0; v < m1->vtxCount; v++)
	{
		p = m0->PointSurfaceDepth(&m1->vtx[v], &depth);
		if ((p > 0) && (depth < best_depth))
		{
			best_obj = 1;
			best_depth = depth;
			VectorCopy(&m1->vtx[v], &best_point);
			VectorCopy(&m0->plane[p].normal, &best_normal);
			collided = true;
			vertex_face = true;
		}
	}

	//test object 0 edge to object 1 edge
	for (int e0 = 0; e0 < m0->edgeCount; e0++)
	{
		for (int e1 = 0; e1 < m1->edgeCount; e1++)
		{
			//avoid excessive ctors
			static line_t del0, del1;
			static vector_t param;
			static vector_t vec0, vec1, cross;

			VectorCopy(&m0->vtx[m0->edge[e0].v[0]], &del0.pos);
			VectorSub(&m0->vtx[m0->edge[e0].v[1]], &m0->vtx[m0->edge[e0].v[0]], &del0.dir);

			VectorCopy(&m1->vtx[m1->edge[e1].v[0]], &del1.pos);
			VectorSub(&m1->vtx[m1->edge[e1].v[1]], &m1->vtx[m1->edge[e1].v[0]], &del1.dir);

			LineLineClosestPointParameter(&del0, &del1, &param.x, &param.y);

			if (param.x < 0 || param.x > 1 || param.y < 0 || param.y > 1)
				continue;

			VectorMultAdd(&del0.pos, &del0.dir, param.x, &vec0);
			VectorMultAdd(&del1.pos, &del1.dir, param.y, &vec1);

			VectorCross(&del0.dir, &del1.dir, &cross);
			if (VectorLengthSq(&cross) < 0.000000001) continue;	//assume parallel
			VectorUnit(&cross, &cross);

			scalar_t depth0 = fInf;
			scalar_t depth1 = fInf;

			int plane0 = m0->PointSurfaceDepth(&vec1, &depth0);
			int plane1 = m1->PointSurfaceDepth(&vec0, &depth1);

			if (plane0 == -1 || plane1 == -1) continue;		//shouldnt ever happen

			//should this be comparing best_depth or only plane_epsilon?  no worries...
			if (depth0 < best_depth && depth1 < best_depth)
			{

#ifdef PHYSICS_PARTICLE_SYSTEM_DEBUG

				particleSystem->CreateParticle(vec0, vectorZero, 1, 0xFFFF00FF, 100, true);
				particleSystem->CreateParticle(vec1, vectorZero, 1, 0xFFFFFF00, 100, true);

				for (scalar_t u = 0; u <= 1; u += 0.1f)
				{
					particleSystem->CreateParticle(del0.pos + del0.dir * u, vectorZero, 2, 0xFF0000FF, 100, true);
					particleSystem->CreateParticle(del1.pos + del1.dir * u, vectorZero, 2, 0xFF00FF00, 100, true);
				}
#endif

				best_obj = (VectorDot(&cross, &o_delta) < 0);		//best_obj is where cross is pointing to
				best_depth = min(depth0,depth1);
				VectorAdd(&vec0, &vec1, &best_point);
				VectorScale(&best_point, 0.5f, &best_point);
				VectorCopy(&cross, &best_normal);
				VectorCopy(&del0.dir, &best_edge0);		//edge0 vector
				VectorCopy(&del1.dir, &best_edge1);		//edge1 vector
				collided = true;
				vertex_face = false;
			}
		}
	}

//	main_console->Print("bestdepth: %f\n", best_depth);

//begin debug
#ifdef PHYSICS_PARTICLE_SYSTEM_DEBUG
	if (collided && !vertex_face)
	{
		particleSystem->CreateParticle(best_point, vectorZero, 2, 0xFF0000FF, 100, true);
		particleSystem->CreateParticle(best_point + best_normal * 1.0f, vectorZero, 0.5, 0xFFFFFF00, 100, true);
		particleSystem->CreateParticle(best_point + best_normal * 2.0f, vectorZero, 0.5, 0xFFFFFF00, 100, true);
		particleSystem->CreateParticle(best_point + best_normal * 3.0f, vectorZero, 0.5, 0xFFFFFF00, 100, true);
		particleSystem->CreateParticle(best_point + best_normal * 4.0f, vectorZero, 0.5, 0xFFFFFF00, 100, true);
	}
#endif
//end debug

	if (!collided)
	{
//		main_console->Print("returning CS_FREE\n");
		return CS_FREE;
	}

	if (best_depth < -planeEpsilon) state = CS_PENETRATING;
	else if (best_depth < planeEpsilon) state = CS_COLLIDING;

	Collision result_info;// = new Collision();

	//store the following:
	result_info.vertex_face = vertex_face;
	VectorCopy(&best_normal, &result_info.normal);
	VectorCopy(&best_point, &result_info.point);
	result_info.obj[0] = o[best_obj];		//object holding vertex
	result_info.obj[1] = o[!best_obj];		//object holding plane

	//if edge edge the save the extra edge/edge info needed for resting collisions
	if (!vertex_face)
	{
		VectorUnit(&best_edge0, &result_info.edge[0]);
		VectorUnit(&best_edge1, &result_info.edge[1]);
	}

//	if (!vertex_face)	//then ensure the direction of our cross vector
//		main_console->Print("collision EDGE_EDGE\n");
//	else
//		main_console->Print("collision VERTEX_FACE\n");
//	main_console->Print("returning state %d\n", state);

	//collisionChain.AddLast(result_info);
	collisionList.push_back(result_info);

	return state;
}

/*
RigidBody
	read:	
	write:	
child functions:
	Collision.Colliding
	Collision.ApplyImpulses
*/
/*
ERROR
when the first collision on the stack is resolved, velocity and rotation are affected
thus obscuring the calculations for all proceeding collisions
*/
void M_Physics::ResolveAllCollisions(void)
{
	int collided, i, count;
	count = collisionList.size();//collisionChain.Count();	//too many C's
	Collision *collision;

	int counter = 0;

	do {
		collided = 0;

		for (i = 0; i < count; i++)
		{
			collision = &collisionList[i];//(Collision *)collisionChain[i];

			if (collision->Colliding(velEpsilon) == COL_COLLIDING)
			{
				//store & accum all forces

				collision->ApplyImpulses();
				collided = 1;

				//tell the solver we had a collision (set the discontinuous flag) *here*
			}
		}

		counter++;

		if (counter > 1000)
		{
			main_console->Print("M_Physics::ResolveAllCollisions(): infinite loop\n");
			return;
		}

	} while (collided);
}

// ********************* THE COLLISION CLASS ********************* 

/*
RigidBody
	read:	x
	write:	
child functions:
	RigidBody.VelocityAtOffset
*/
int Collision::Colliding(scalar_t velEpsilon)
{
	vector_t	veldif;
	vector_t	offset[2];
	scalar_t	relative_vel;

	VectorSub(&point, &obj[0]->x, &offset[0]);
	VectorSub(&point, &obj[1]->x, &offset[1]);

	vector_t v0, v1;
	obj[0]->VelocityAtOffset(&offset[0], &v0);
	obj[1]->VelocityAtOffset(&offset[1], &v1);
	VectorSub(&v0, &v1, &veldif);

	relative_vel = VectorDot(&veldif, &normal);

	if (relative_vel > velEpsilon) return COL_FREE;
	if (relative_vel > -velEpsilon) return COL_RESTING;

	return COL_COLLIDING;
}

/*
RigidBody
	read:	P, L, SurfaceRestitution, MassInv, InertiaInv
	write:	P, L
child functions:
	RigidBody.VelocityAtOffset
	RigidBody.ComputeAuxiliary
*/
/*
void calc_collision_forces(obj_t *o0, obj_t *o1, vec2_t s_intersection, vec2_t s_plane_normal)

applies impulse force to o0 and negative impulse force to o1

o0 = object to which the force is being applied to
o1 = object to which negative force is being applied to
s_intersection = point of collision between objects
s_plane_normal = unit normal of the plane on which the collision exists
*/
void Collision::ApplyImpulses(void)
{
	RigidBody *o0, *o1;
	vector_t offset[2];
	vector_t vel_ab;
	vector_t impulse_force;
	scalar_t restitution;

	vector_t reg;		//temp vector register

	o0 = obj[0];
	o1 = obj[1];

#ifdef PHYSICS_PARTICLE_SYSTEM_DEBUG
	static int displayCollision = 1;
	static M_CVar *cvar = new M_CVar(&displayCollision, "DisplayCollision");
	if (displayCollision) particleSystem->CreateParticle(point, vectorZero, 5, 0xFFFFFFFF, 150, true);
#endif

//calc collision offset
	VectorSub(&point, &obj[0]->x, &offset[0]);
	VectorSub(&point, &obj[1]->x, &offset[1]);

////grab dif in velocities (incorperate addition via angular velocity)
	obj[0]->VelocityAtOffset(&offset[0], &vel_ab);
	obj[1]->VelocityAtOffset(&offset[1], &reg);
	VectorSub(&vel_ab, &reg, &vel_ab);

	restitution = obj[0]->SurfaceRestitution * obj[1]->SurfaceRestitution;

//calc impulse force
//             _      _    _   _                      _      _           _      _
//j = -(1 + e) v_ab * n / (n * n * (1/m_a + 1/m_b) + (r_ap x n)^2/I_a + (r_bp x n)^2/I_b)
	scalar_t numerator = -(1.f + restitution) * VectorDot(&vel_ab, &normal);

	scalar_t denominator = o0->MassInv + o1->MassInv;		//scale this value by N*N if the collision normal is not a unit vector
//angular momentum influence
	VectorCross(&offset[0], &normal, &reg);
	MatrixMul(&o0->InertiaInv, &reg, &reg);
	VectorCross(&reg, &offset[0], &reg);
	denominator += VectorDot(&normal, &reg);
	//denominator += normal % ((o0->InertiaInv * (offset[0] * normal)) * offset[0]);
	VectorCross(&offset[1], &normal, &reg);
	MatrixMul(&o1->InertiaInv, &reg, &reg);
	VectorCross(&reg, &offset[1], &reg);
	denominator += VectorDot(&normal, &reg);
	//denominator += normal % ((o1->InertiaInv * (offset[1] * normal)) * offset[1]);

	scalar_t j = numerator / denominator;

//create force vector from force magnitude
	VectorScale(&normal, j, &impulse_force);

//apply impulses - directly to velocity, as to not cause d/dt errors

	VectorAdd(&o0->P, &impulse_force, &o0->P);
	VectorSub(&o1->P, &impulse_force, &o1->P);

//angular momentum influence
	VectorCross(&offset[0], &impulse_force, &reg);
	VectorAdd(&o0->L, &reg, &o0->L);
	VectorCross(&offset[1], &impulse_force, &reg);
	VectorSub(&o1->L, &reg, &o1->L);

	//recompute auxiliary vars
	o0->ComputeAuxiliary();
	o1->ComputeAuxiliary();
}





/* bigvector_t / bigmatrix_t interface */

//typedef struct {
//	int			dim;
//	scalar_t	v[1];
//} bigvector_t;

class BigVector {
	int h;
	scalar_t *v;
public:
	BigVector(int h) {
		this->h = h;
		v = new scalar_t[h];
		assert(v);
	}
	~BigVector() {
		assert(v);
		delete v;
	}

	scalar_t &m(int row) {
		assert(v);
		assert(row >= 0 && row < h);
		return v[row];
	}

	int rows(void) const { return h; }
};

//bigvector_t *CreateBigVector(int n)
//{
//	bigvector_t *v = (bigvector_t *)malloc(sizeof(int) + sizeof(scalar_t) * n);
//	if (!v) return NULL;
//
//	v->dim = n;
//	return v;
//}
//
//void DestroyBigVector(bigvector_t *v)
//{
//	free(v);
//}

class BigMatrix {
	int w, h;
	scalar_t *v;
public:
	BigMatrix(int w, int h) {
		this->w = w;
		this->h = h;
		v = new scalar_t[w*h];
		assert(v);
	}

	~BigMatrix() {
		assert(v);
		delete v;
	}

	scalar_t &m(int row, int col) {
		assert(v);
		assert(row >= 0 && row < h);
		assert(col >= 0 && col < w);
		return v[col+row*w];
	}

	int rows(void) const { return h; }
	int cols(void) const { return w; }
};

//typedef struct {
//	int height, width;
//	scalar_t v[1];
//} bigmatrix_t;

//bigmatrix_t *CreateBigMatrix(int h, int w)
//{
//	bigmatrix_t *m = (bigmatrix_t *)malloc(sizeof(int)*2 + sizeof(scalar_t) *w*h);
//	if (!m) return NULL;
//
//	m->height = h;
//	m->width = w;
//	return m;
//}
//
//void DestroyBigMatrix(bigmatrix_t *m)
//{
//	free(m);
//}

//and the goods

void Compute_Nderiv(Collision *c, vector_t *dN_dt)
{
	if (c->vertex_face)
	{
		VectorCross(&c->obj[1]->omega, &c->normal, dN_dt);
		return;
	}
	else
	{
		vector_t edot[2], z,n,reg;
		scalar_t ilen;

		//compute 'n'
		VectorCross(&c->edge[0], &c->edge[1], &n);
		ilen = VectorLength(&n);
		if (ilen) ilen = 1.f / ilen;
		VectorScale(&n, ilen, &n);

		VectorCross(&c->obj[0]->omega, &c->edge[0], &edot[0]);
		VectorCross(&c->obj[1]->omega, &c->edge[1], &edot[1]);

		//compute 'z'
		VectorCross(&edot[0], &c->edge[1], &z);
		VectorCross(&edot[1], &c->edge[0], &reg);
		VectorAdd(&z, &reg, &z);

		VectorCross(&z, &c->normal, &reg);
		VectorCross(&reg, &c->normal, &reg);
		VectorSub(&z, &reg, &z);
		VectorScale(&z, ilen, dN_dt);
	}
}

scalar_t Compute_Aij(Collision *ci, Collision *cj)
{
	/* If the bodies involved in the ith and jth contact are
	distinct, then ai j is zero. */
	if ((ci->obj[0] != cj->obj[0]) &&
		(ci->obj[1] != cj->obj[1]) &&
		(ci->obj[0] != cj->obj[1]) &&
		(ci->obj[1] != cj->obj[0]))
		return 0.0;

	vector_t offset[2];

	VectorSub(&ci->point, &ci->obj[0]->x, &offset[0]);
	VectorSub(&ci->point, &ci->obj[1]->x, &offset[1]);

	/* What force and torque does contact j exert on body ci->obj[0]? */
	vector_t force_on_a = {0,0,0};
	vector_t torque_on_a = {0,0,0};

	if(cj->obj[0] == ci->obj[0])
	{
		/* force direction of jth contact force on ci->obj[0] */
		VectorCopy(&cj->normal, &force_on_a);
		/* torque direction */
		VectorSub(&cj->point, &ci->obj[0]->x, &torque_on_a);
		VectorCross(&torque_on_a, &cj->normal, &torque_on_a);
	}
	else if(cj->obj[1] == ci->obj[0])
	{
		VectorScale(&cj->normal, -1, &force_on_a);
		VectorSub(&cj->point, &ci->obj[0]->x, &torque_on_a);
		VectorCross(&torque_on_a, &cj->normal, &torque_on_a);
	}

	/* What force and torque does contact j exert on body ci->obj[1]? */
	vector_t force_on_b = {0,0,0};
	vector_t torque_on_b = {0,0,0};

	if(cj->obj[0] == ci->obj[1])
	{
		/* force direction of jth contact force on ci->obj[1] */
		VectorCopy(&cj->normal, &force_on_b);
		/* torque direction */
		VectorSub(&cj->point, &ci->obj[1]->x, &torque_on_b);
		VectorCross(&torque_on_b, &cj->normal, &torque_on_b);
	}
	else if(cj->obj[1] == ci->obj[1])
	{
		/* force direction of jth contact force on ci->obj[1] */
		VectorScale(&cj->normal, -1, &force_on_b);
		/* torque direction */
		VectorSub(&cj->point, &ci->obj[1]->x, &torque_on_b);
		VectorCross(&torque_on_b, &cj->normal, &torque_on_b);
	}

	/* Now compute how the jth contact force affects the linear
	and angular acceleration of the contact point on body ci->obj[0] */
	vector_t a_linear;
	VectorScale(&force_on_a, ci->obj[0]->MassInv, &a_linear);

	vector_t a_angular;
	MatrixMul(&ci->obj[0]->InertiaInv, &torque_on_a, &a_angular);
	VectorCross(&a_angular, &offset[0], &a_angular);

	/* Same for ci->obj[1] */
	vector_t b_linear;
	VectorScale(&force_on_b, ci->obj[1]->MassInv, &b_linear);

	vector_t b_angular;
	MatrixMul(&ci->obj[1]->InertiaInv, &torque_on_b, &b_angular);
	VectorCross(&b_angular, &offset[1], &b_angular);

	vector_t sum;
	VectorAdd(&a_linear, &a_angular, &sum);
	VectorSub(&sum, &b_linear, &sum);
	VectorSub(&sum, &b_angular, &sum);

	return VectorDot(&ci->normal, &sum);
}


//accepts collisoin chain
BigMatrix *Compute_A_Matrix(std::vector<Collision> &list/*LinkedList *chain*/, int count)
{
	BigMatrix *a = new BigMatrix(count,count);
	if (!a) return NULL;

//	int i = 0;
//	for (Collision *ci = (Collision *)chain->head; ci; ci = (Collision *)ci->chain_next)
	for (int i = 0; i < list.size(); i++) 
	{
		Collision *ci = &list[i];
//		int j = 0;
//		for (Collision *cj = (Collision *)chain->head; cj; cj = (Collision *)cj->chain_next)
		for (int j = 0; j < list.size(); j++)
		{
			Collision *cj = &list[j];
			//i = Y
			//j = X
			//store [height][width] so we can just dot vectors for matrix mul
			//a[i,j] =...
			a->m(i,j) = Compute_Aij(ci, cj);
//			j++;
		}
//		i++;
	}

	return a;
}

//accepts collisoin chain
BigVector *Compute_B_Vector(std::vector<Collision> &list/*LinkedList *chain*/, int count)
{
	int i,j;

	BigVector *b = new BigVector(count);
	if (!b) return NULL;

//	i = 0;
//	for (Collision *c = (Collision *)chain->head; c; c = (Collision *)c->chain_next, i++)
	for (i = 0; i < list.size(); i++)
	{
		Collision *c = &list[i];
		vector_t reg;

		vector_t offset[2];
		VectorSub(&c->point, &c->obj[0]->x, &offset[0]);
		VectorSub(&c->point, &c->obj[1]->x, &offset[1]);

		vector_t ext_part[2];
		vector_t vel_part[2];

		for (j = 0; j < 2; j++)
		{
			//compute external force/torque aspect
			MatrixMul(&c->obj[j]->InertiaInv, &c->obj[j]->torque, &ext_part[j]);
			VectorCross(&ext_part[j], &offset[j], &ext_part[j]);
			VectorMultAdd(&ext_part[j], &c->obj[j]->force, c->obj[j]->MassInv, &ext_part[j]);

			//compute velocity aspect
			VectorCross(&c->obj[j]->omega, &offset[j], &vel_part[j]);
			VectorCross(&c->obj[j]->omega, &vel_part[j], &vel_part[j]);

			VectorCross(&c->obj[j]->L, &c->obj[j]->omega, &reg);
			MatrixMul(&c->obj[j]->InertiaInv, &reg, &reg);
			VectorCross(&reg, &offset[j], &reg);

			VectorAdd(&vel_part[j], &reg, &vel_part[j]);
		}

		VectorAdd(&ext_part[0], &vel_part[0], &reg);
		VectorSub(&reg, &ext_part[1], &reg);
		VectorSub(&reg, &vel_part[1], &reg);
		scalar_t k1 = VectorDot(&c->normal, &reg);

		//compute relative velocity
		vector_t vrel;
		c->obj[0]->VelocityAtOffset(&offset[0], &vrel);
		c->obj[1]->VelocityAtOffset(&offset[1], &reg);
		VectorSub(&vrel, &reg, &vrel);

		//compute dN/dt
		vector_t ndot;
		Compute_Nderiv(c, &ndot);

		scalar_t k2 = 2 * VectorDot(&ndot, &vrel);

		b->m(i) = k1 + k2;
	}

	return b;
}

//notice - thrashes both =)
BigVector *QP_Solve(BigMatrix *a, BigVector *b)
{
//	int i,j,k;
//	//Ax = b
//	//if there is no solution for b then find a b' within colspace(A) and solve for x
//	//b' = proj_w(b)
//	//Ax = b
//	//AtAx = Atb
//	//x = (At*A)-1*At*b
//
//	//first print ax=b:
//#ifdef DEBUG_PRINT
//	main_console->Print("matrix eq\n");
//	for (i = 0; i < a->width; i++)
//	{
//		for (j = 0; j < a->height; j++)
//		{
//			main_console->Print("%f ", a->v[j+i*a->width]);
//		}
//		main_console->Print(" = %f\n", b->v[i]);
//	}
//#endif
//
//	//row reduce...
//	BigVector *x = new BigVector(a->rows());
//	if (!x) return NULL;
//
//	//scan down the diagonal... (its a square matrix =) )
//	for (i = 0; i < a->cols(); i++)
//	{
//		//first set a[i,i] to 1
//		//do so by scaling the i'th row
//		scalar_t s = a->m(i,i);
//		if (s == 0) {
//			*main_console << "swapping needed\n";
//			return NULL;
//		} else {
//			//scale the row - set a '1' in the diagonal
//			s = 1.f / s;
//			//since were making an upper diagonal matrix, just start at 'i' (all else will be zeros)
//			for (j = i; j < a->width; j++)
//				a->m(i,j) *= s;
//			//and dont forget to manip the 'b'
//			b->m(i) *= s;
//		}
//
//		//now cycle down the height below i and zero all columns
//		for (j = i+1; j < a->rows(); j++)
//		{
//			//row j = row j - row i * a[j,i]
//			s = a->m(j,i)//i'th element in the row
//			//since were making an upper diagonal matrix, just start at 'i' (all else will be zeros)
//			for (k = i; k < a->width; k++)
//			{
//				a->m(j,k) -= a->m(i,k) * s;
//			}
//			//and dont forget about bob
//			b->m(i) -= b->m(i) * s;
//		}
//
//	}
//
//	//first print ax=b:
//#ifdef DEBUG_PRINT
//	main_console->Print("row reduced:\n");
//	for (i = 0; i < a->cols(); i++)
//	{
//		for (j = 0; j < a->rows(); j++)
//		{
//			main_console->Print("%f ", a->m(j,i));
//		}
//		main_console->Print(" = %f\n", b->v[i]);
//	}
//#endif
//
//	//and now we have an upper triangular matrix =)
//	//with 1's along the diagonal
//#ifdef DEBUG_PRINT
//	main_console->Print("solution:\n");
//#endif
//	for (i = a->width-1; i >= 0; i--)
//	{
//		//start with the solution in 'b'
//		x->v[i] = b->v[i];
//		//backwards substitute all previously calculated 'x's
//		for (j = i+1; j < a->width; j++)
//		{
//			x->v[i] -= a->v[j+i*a->width] * x->v[j];
//		}
//		//scale by diagonal (should be '1' anyways)
//		if (a->v[i+i*a->width])
//		{
//			x->v[i] /= a->v[i+i*a->width];
//		}
//#ifdef DEBUG_PRINT
//		//and we should be good to go
//		main_console->Print("%f\n", x->v[i]);
//#endif
//	}
//
//
//	return x;

	BigVector *x = new BigVector(a->rows());
	BigMatrix *v = new BigMatrix(a->rows(), a->cols());
	if (!x) return NULL;

	for (int k = 0; k < a->cols(); k++) {
		scalar_t len;
		int j;
		//set x and v
		for (j = k; j < a->rows(); j++) {
			int j_k = j - k;
			scalar_t ajk = a->m(j,k);
			x->m(j_k) = ajk;
			v->m(j_k,k) = ajk;
			len += ajk * ajk;
		}
		len = (scalar_t)sqrt(len);

		//add x[0] onto v[0]
		if (x->m(0) < 0) {
			v->m(0,k) -= len;
		} else {
			v->m(0,k) += len;
		}

		//normalize v
		len = 0;
		for (int j = 0; j < a->rows()-k; j++) {
			len += v->m(j,k) * v->m(j,k);
		}
		if (len) len = (scalar_t)(1.0 / sqrt(len));
		for (int j = 0; j < a->rows()-k; j++) {
			v->m(j,k) *= len;
		}


	}
	return x;
}

void M_Physics::ComputeContactForces(void)
{
	int count = collisionList.size();//collisionChain.Count();	//save some computations
	if (!count) return;

	Collision *c;
	//only talk to the collision objects that are RESTING
	//first - filter out all the non-resting collisions
//	for (c = (Collision *)collisionChain.head; c; c = next)
	for (int i =0; i < collisionList.size(); i++) {
		Collision *c = &collisionList[i];
//		next = (Collision *)c->chain_next;
		//if not resting then delete
		if (c->Colliding(velEpsilon) != COL_RESTING) {

			//
//			delete c;
			//
//			collisionList.erase(collisionList.begin() + i);
			//
			for (int j = i+1; j < collisionList.size(); j++) {
				collisionList[j-1] = collisionList[j];
			}
			collisionList.resize(collisionList.size()-1);

			i--;
		}
	}

	count = collisionList.size();//collisionChain.Count();	//save some computations
	if (!count) return;

	main_console->Print("contact force count %d\n", count);

	//next - solve.
	BigMatrix *amat = Compute_A_Matrix(collisionList, count);
	if (!amat) return;

	BigVector *bvec = Compute_B_Vector(collisionList, count);
	if (!bvec)
	{
		delete amat;
		return;
	}

	BigVector *fvec = QP_Solve(amat, bvec);

	delete amat;
	delete bvec;

	if (!fvec) return;

	for (i = 0; i < count; i++) {
		c = &collisionList[i];//(Collision *)collisionChain[i];
		RigidBody *o0 = c->obj[0];
		RigidBody *o1 = c->obj[1];

		vector_t force, reg;
		VectorScale(&c->normal, fvec->m(i), &force);

		VectorAdd(&o0->force, &force, &o0->force);
		VectorSub(&o1->force, &force, &o1->force);

		VectorSub(&c->point, &o0->x, &reg);
		VectorCross(&reg, &force, &reg);
		VectorAdd(&o0->torque, &reg, &o0->torque);

		VectorSub(&c->point, &o1->x, &reg);
		VectorCross(&reg, &force, &reg);
		VectorSub(&o1->torque, &reg, &o1->torque);
	}

	delete fvec;
}