/*

    File: Physics.h

    Copyright (C) 2000-2007 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#ifndef PHYSICS_H
#define PHYSICS_H

#include <vector>

#include "../m_math3d/matrix.h"
#include "../m_math3d/vector.h"
#include "../m_foundation/chain.h"
#include "../m_foundation/const.h"
#include "cmodel.h"

class RigidBody {
public:
	RigidBody();
	~RigidBody();

//constant quantities
	scalar_t MassInv;			//default 0 (infinite mass, no momentum change)
	matrix_t BodyInertiaInv;	//default 0 (infinite inertia, no rotational change)
//state variables
	vector_t x;					//position, default (0,0,0)
	quat_t	 q;					//rotation, default (0,0,0)
	vector_t P;					//momentum, default (0,0,0)
	vector_t L;					//angular momentum, default (0,0,0)
//derived quantities
	matrix_t InertiaInv;		//world oriented inverse inertia matrix
	vector_t v;					//velocity
	vector_t omega;				//rotation
//computed quantities
	vector_t force;				//force
	vector_t torque;			//torque

	CModel *model;			//the collision model, default null (void of collision test)
	vector_t size;			//the scaled size of the object / collision model, default 1
	scalar_t densityInv;		//the inverse density of the model.  used to compute mass / inertia

	//this value could instead be made a cmodel collision face property...
	scalar_t SurfaceRestitution;	//surface collision restutition, default 1

	bool integrate;		//whether or not we can integrate the object (movement or not)		- default true
	bool collides_solid;//whether or not this object collides with other solid objects		- default true
	bool is_solid;		//whether or not this object accepts collisions from other objects	- default true

	void CalcMassInertia(void);			//compute MassInv,BodyInertiaInv based upon model,size,densityInv
	void ComputeInertiaInv(void);		//compute InertiaInv based upon q
	void ComputeAuxiliary(void);		//compute v, omega based upon P,L
	void VelocityAtOffset(const vector_t *offset, vector_t *vres);
};

//ONLY USED FOR CONTACT/COLLISION FORCES

class Collision {
public:
	RigidBody	*obj[2];		//the two objects of collision.  o0 is assumed to hold the vertex, o1 the plane

	bool		vertex_face;	//set to 'true' if vertex/face contact is being used

	vector_t	point;			//point of collision.  assumed to be a vertex of o0
	vector_t	normal;			//collision plane normal.  assumed to be pointing from o1 towards o0
	vector_t	edge[2];		//used for edge on edge collision.  the normal vectors of the 2 edges

#define COL_FREE		0
#define COL_RESTING		1
#define COL_COLLIDING	2
	int			Colliding(scalar_t vel_epsilon);

	void		ApplyImpulses(void);

	//MSVC5 std::vector fix:
	bool operator<(const Collision &c) const { return &c < this; }
	bool operator==(const Collision &c) const { return &c == this; }
};

class M_Physics {
public:
	M_Physics() {Reset();}
	~M_Physics() {Kill();}

	bool Init(void /*int max*/);	//init
	void Reset(void);				//clear all vars
	void Kill(void);				//shutdown

	void Update(double delta_time);
	void Display(void);

	//add a RigidBody to the Linked List
	void AddRigidBody(RigidBody *body);

	//removes a RigidBody from the LinkedList - ALWAYS CALL THIS BEFORE YOU DELETE A RIGID BODY!!!
	void RemoveRigidBody(RigidBody *body);

	//sets the callback force function for each body
	void SetForceFunction(void (*func)(M_Physics *));

	//The distance at which objects are to be considered touching.  public 'cuz im lazy.
	scalar_t planeEpsilon;
	scalar_t velEpsilon;
	int useContactForces;

	//external access:

	int getBodyCount(void) {
		return rigidBodyArray.size();
	}

	RigidBody *getBody(int index) { 
		return (index >= 0 && index < getBodyCount()) ? rigidBodyArray[index] : NULL;
	}

private:
	//our physics engine will contain...
	//collisions
//	LinkedList collisionChain;				//collisionchain handled internally 
	std::vector<Collision> collisionList;

	std::vector<RigidBody *> rigidBodyArray;

	//the force function - called on Update() w/each body passed as the param
	//this is for the user to specify any environmental forces acting on the body
	//eg. viscousity, gravity, magnetism, whatever
	void	(*forceFunc)(M_Physics *);

	void ApplyForces(void);
	void UpdateSystem(double delta_time);
	void Integrate(double delta_time);

	int CheckCollisionState(void);
	int TestCollision(CModel *m0, CModel *m1, RigidBody *o0, RigidBody *o1);

	void ResolveAllCollisions(void);

//from here on out is the resting collision feature
	void ComputeContactForces(void);
};

#endif